#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Utilities"""

import logging
from typing import Any, Literal

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
import numpy.typing as npt
import optimistix as optx
from jax import Array, lax
from jax.tree_util import tree_map
from jax.typing import ArrayLike
from scipy.constants import kilo, mega

from atmodeller import max_exp_input
from atmodeller.constants import ATMOSPHERE, BOLTZMANN_CONSTANT_BAR, OCEAN_MASS_H2

logger: logging.Logger = logging.getLogger(__name__)

OptxSolver = optx.AbstractRootFinder | optx.AbstractLeastSquaresSolver | optx.AbstractMinimiser


@eqx.filter_jit
def get_log_number_density_from_log_pressure(
    log_pressure: ArrayLike, temperature: ArrayLike
) -> Array:
    """Gets log number density from log pressure

    Args:
        log_pressure: Log pressure
        temperature: Temperature

    Returns:
        Log number density
    """
    log_number_density: Array = (
        -jnp.log(BOLTZMANN_CONSTANT_BAR) - jnp.log(temperature) + log_pressure
    )

    return log_number_density


@eqx.filter_jit
def safe_exp(x: ArrayLike) -> Array:
    return jnp.exp(jnp.clip(x, a_max=max_exp_input))


@eqx.filter_jit
def partial_rref_jax(matrix: Array) -> Array:
    """Computes the partial reduced row echelon form to determine linear components.

    This function is currently not used, since the reduction of matrices is performed prior to
    JAX operations. But since this function took some time to figure out, it is retained for
    potential future use. Note that lax.fori_loop is not compatible with reverse differentiation

    Args:
        matrix: The matrix to compute the reduced row echelon form

    Returns:
        A matrix of linear components.
    """
    nrows, ncols = matrix.shape
    augmented_matrix: Array = jnp.hstack((matrix, jnp.eye(nrows)))

    def swap_rows(matrix: Array, row1: int, row2: int) -> Array:
        """Swaps two rows in a matrix.

        Args:
            matrix: Matrix
            row1: Row1 index
            row2: Row2 index

        Returns:
            Matrix with the rows swapped
        """
        row1_data: Array = lax.dynamic_slice(matrix, (row1, 0), (1, matrix.shape[1]))
        row2_data: Array = lax.dynamic_slice(matrix, (row2, 0), (1, matrix.shape[1]))
        matrix = lax.dynamic_update_slice(matrix, row2_data, (row1, 0))
        matrix = lax.dynamic_update_slice(matrix, row1_data, (row2, 0))
        return matrix

    def find_nonzero_row(matrix: Array, i: int) -> int:
        """Finds the first non-zero element in the column below the pivot.

        Args:
            matrix: Matrix
            i: Row index

        Returns:
            Relative row index of the first non-zero element below row i
        """

        def body_fun(j: int, nonzero_row: int) -> int:
            """Body function

            Args:
                j: Row offset from row i
                nonzero_row: Non-zero row index found

            Returns:
                The minimum row offset with a non-zero element
            """
            value: Array = lax.dynamic_slice(matrix, (i + j, i), (1, 1))[0, 0]
            # nonzero_row == -1 indicates that no non-zero element has been found yet
            return lax.cond(
                (value != 0) & (nonzero_row == -1),
                lambda _: j,
                lambda _: nonzero_row,
                operand=None,
            )

        nonzero_row: Array = lax.fori_loop(0, nrows - i, body_fun, -1)
        return lax.cond(nonzero_row == -1, lambda _: i, lambda _: nonzero_row + i, operand=None)

    def forward_step(i: int, matrix: Array) -> Array:
        """Forward step

        Args:
            i: Current row
            matrix: Matrix

        Returns:
            Matrix
        """
        # Check if the pivot element is zero and swap rows to get a non-zero pivot element.
        pivot_value: Array = lax.dynamic_slice(matrix, (i, i), (1, 1))[0, 0]
        # jax.debug.print("pivot_value = {out}", out=pivot_value)
        nonzero_row: int = lax.cond(
            pivot_value == 0, lambda _: find_nonzero_row(matrix, i), lambda _: i, operand=None
        )
        matrix = lax.cond(
            nonzero_row != i,
            lambda _: swap_rows(matrix, i, nonzero_row),
            lambda _: matrix,
            operand=None,
        )

        def eliminate_below_row(j: int, matrix: Array) -> Array:
            """Eliminates below the row

            Args:
                j: Row offset from row i
                matrix: Matrix

            Returns:
                Matrix
            """
            pivot: Array = lax.dynamic_slice(matrix, (i, i), (1, 1))[0, 0]
            ratio: Array = lax.dynamic_slice(matrix, (j, i), (1, 1))[0, 0] / pivot
            row_i: Array = lax.dynamic_slice(matrix, (i, 0), (1, ncols + nrows))
            row_j: Array = lax.dynamic_slice(matrix, (j, 0), (1, ncols + nrows))
            return lax.dynamic_update_slice(matrix, row_j - ratio * row_i, (j, 0))

        def loop_body(j: int, matrix: Array) -> Array:
            return eliminate_below_row(j, matrix)

        matrix = lax.fori_loop(i + 1, nrows, loop_body, matrix)

        return matrix

    def backward_step(i: int, matrix: Array) -> Array:
        """Backward step

        Args:
            i: Current row
            matrix: Matrix

        Returns:
            Matrix
        """
        # Normalize the pivot row.
        pivot: Array = lax.dynamic_slice(matrix, (i, i), (1, 1))[0, 0]
        normalized_row = lax.dynamic_slice(matrix, (i, 0), (1, ncols + nrows)) / pivot
        matrix = lax.dynamic_update_slice(matrix, normalized_row, (i, 0))

        def eliminate_above_row(j: int, matrix: Array) -> Array:
            """Eliminates above the row

            Args:
                j: Row offset from row i
                matrix: Matrix

            Returns:
                Matrix
            """
            is_nonzero: Array = lax.dynamic_slice(matrix, (j, i), (1, 1))[0, 0] != 0

            def eliminate_row(matrix: Array) -> Array:
                ratio: Array = lax.dynamic_slice(matrix, (j, i), (1, 1))[0, 0] / pivot
                row_i: Array = lax.dynamic_slice(matrix, (i, 0), (1, ncols + nrows))
                row_j: Array = lax.dynamic_slice(matrix, (j, 0), (1, ncols + nrows))
                return lax.dynamic_update_slice(matrix, row_j - ratio * row_i, (j, 0))

            return lax.cond(is_nonzero, eliminate_row, lambda matrix: matrix, matrix)

        def loop_body(j: int, matrix: Array) -> Array:
            return eliminate_above_row(j, matrix)

        matrix = lax.fori_loop(0, i, loop_body, matrix)

        return matrix

    def forward_elimination_body(i: int, matrix: Array) -> Array:
        return forward_step(i, matrix)

    augmented_matrix = lax.fori_loop(0, ncols, forward_elimination_body, augmented_matrix)

    def backward_elimination_body(i: int, matrix: Array) -> Array:
        return backward_step(ncols - 1 - i, matrix)

    augmented_matrix = lax.fori_loop(0, ncols, backward_elimination_body, augmented_matrix)

    # Don't need the reduced matrix, but maybe useful for debugging
    # reduced_matrix = lax.dynamic_slice(augmented_matrix, (0, 0), (nrows, ncols))
    component_matrix: Array = lax.dynamic_slice(
        augmented_matrix, (ncols, ncols), (nrows - ncols, nrows)
    )

    return component_matrix


def partial_rref(matrix: npt.NDArray) -> npt.NDArray:
    """Computes the partial reduced row echelon form to determine linear components

    Returns:
        A matrix of linear components
    """
    nrows, ncols = matrix.shape

    augmented_matrix: npt.NDArray = np.hstack((matrix, np.eye(nrows)))
    # debug("augmented_matrix = \n%s", augmented_matrix)
    # Permutation matrix
    # P: npt.NDArray = np.eye(nrows)

    # Forward elimination with partial pivoting
    for i in range(ncols):
        # Check if the pivot element is zero and swap rows to get a non-zero pivot element.
        if augmented_matrix[i, i] == 0:
            nonzero_row: int = np.nonzero(augmented_matrix[i:, i])[0][0] + i
            augmented_matrix[[i, nonzero_row], :] = augmented_matrix[[nonzero_row, i], :]
            # P[[i, nonzero_row], :] = P[[nonzero_row, i], :]
        # Perform row operations to eliminate values below the pivot.
        for j in range(i + 1, nrows):
            ratio: float = augmented_matrix[j, i] / augmented_matrix[i, i]
            augmented_matrix[j] -= ratio * augmented_matrix[i]
    # logger.debug("augmented_matrix after forward elimination = \n%s", augmented_matrix)

    # Backward substitution
    for i in range(ncols - 1, -1, -1):
        # Normalize the pivot row.
        augmented_matrix[i] /= augmented_matrix[i, i]
        # Eliminate values above the pivot.
        for j in range(i - 1, -1, -1):
            if augmented_matrix[j, i] != 0:
                ratio = augmented_matrix[j, i] / augmented_matrix[i, i]
                augmented_matrix[j] -= ratio * augmented_matrix[i]
    # logger.debug("augmented_matrix after backward substitution = \n%s", augmented_matrix)

    # reduced_matrix: npt.NDArray = augmented_matrix[:, :ncols]
    component_matrix: npt.NDArray = augmented_matrix[ncols:, ncols:]
    # logger.debug("reduced_matrix = \n%s", reduced_matrix)
    # logger.debug("component_matrix = \n%s", component_matrix)
    # logger.debug("permutation_matrix = \n%s", P)

    return component_matrix


class UnitConversion(eqx.Module):
    """Unit conversions"""

    atmosphere_to_bar: float = ATMOSPHERE
    bar_to_Pa: float = 1.0e5
    bar_to_MPa: float = 1.0e-1
    bar_to_GPa: float = 1.0e-4
    Pa_to_bar: float = 1.0e-5
    MPa_to_bar: float = 1.0e1
    GPa_to_bar: float = 1.0e4
    fraction_to_ppm: float = mega
    g_to_kg: float = 1 / kilo
    ppm_to_fraction: float = 1 / mega
    ppm_to_percent: float = 100 / mega
    percent_to_ppm: float = 1.0e4
    cm3_to_m3: float = 1.0e-6
    m3_to_cm3: float = 1.0e6
    m3_bar_to_J: float = 1.0e5
    J_to_m3_bar: float = 1.0e-5
    litre_to_m3: float = 1.0e-3


unit_conversion = UnitConversion()


def bulk_silicate_earth_abundances() -> dict[str, dict[str, float]]:
    """Bulk silicate Earth element masses in kg.

    Hydrogen, carbon, and nitrogen from :cite:t:`SKG21`
    Sulfur from :cite:t:`H16`
    Chlorine from :cite:t:`KHK17`
    """
    earth_bse: dict[str, dict[str, float]] = {
        "H": {"min": 1.852e20, "max": 1.894e21},
        "C": {"min": 1.767e20, "max": 3.072e21},
        "S": {"min": 8.416e20, "max": 1.052e21},
        "N": {"min": 3.493e18, "max": 1.052e19},
        "Cl": {"min": 7.574e19, "max": 1.431e20},
    }

    for _, values in earth_bse.items():
        values["mean"] = np.mean((values["min"], values["max"]))  # type: ignore

    return earth_bse


def earth_oceans_to_hydrogen_mass(number_of_earth_oceans: ArrayLike = 1) -> ArrayLike:
    """Converts Earth oceans to hydrogen mass

    Args:
        number_of_earth_oceans: Number of Earth oceans. Defaults to 1.

    Returns:
        Hydrogen mass
    """
    h_grams: ArrayLike = number_of_earth_oceans * OCEAN_MASS_H2
    h_kg: ArrayLike = h_grams * unit_conversion.g_to_kg

    return h_kg


class ExperimentalCalibration(eqx.Module):
    """Experimental calibration

    Args:
        temperature_min: Minimum calibrated temperature. Defaults to None.
        temperature_max: Maximum calibrated temperature. Defaults to None.
        pressure_min: Minimum calibrated pressure. Defaults to None.
        pressure_max: Maximum calibrated pressure. Defaults to None.
        log10_fO2_min: Minimum calibrated log10 fO2. Defaults to None.
        log10_fO2_max: Maximum calibrated log10 fO2. Defaults to None.
    """

    temperature_min: float | None = None
    temperature_max: float | None = None
    pressure_min: float | None = None
    pressure_max: float | None = None
    log10_fO2_min: float | None = None
    log10_fO2_max: float | None = None


@eqx.filter_jit
def power_law(values: ArrayLike, constant: ArrayLike, exponent: ArrayLike) -> Array:
    """Power law

    Args:
        values: Values
        constant: Constant for the power law
        exponent: Exponent for the power law

    Returns:
        Evaluated power law
    """
    return jnp.power(values, exponent) * constant


def is_arraylike_batched(x: ArrayLike) -> Literal[0, None]:
    """Checks if x is batched.

    Args:
        x: Something arraylike

    Returns:
        0 (axis) if batched, else None (not batched)
    """
    return 0 if eqx.is_array(x) and x.ndim > 0 else None  # type: ignore


def vmap_axes_spec(x: Any) -> Any:
    """Recursively generate in_axes for vmap by checking if each leaf is batched (axis 0).

    Args:
        x: Pytree of nested containers possibly containing arrays or scalars

    Returns:
        Pytree matching the structure of x
    """
    return tree_map(is_arraylike_batched, x)


def get_batch_size(x: Any) -> int:
    """Determines the maximum batch size (i.e., length along axis 0) among all array-like leaves.

    Args:
        x: Pytree of nested containers possibly containing arrays or scalars

    Returns:
        The maximum size along axis 0 among all array-like leaves, or 0 if all leaves are scalars.
    """
    max_size: int = 0
    for leaf in jax.tree_util.tree_leaves(x):
        if eqx.is_array(leaf) and leaf.ndim > 0:
            max_size = max(max_size, leaf.shape[0])

    return max_size


def pytree_debug(pytree: Any, name: str) -> None:
    """Prints the pytree structure for debugging vmap.

    Args:
        pytree: Pytree to print
        name: Name for the debug print
    """
    arrays, static = eqx.partition(pytree, eqx.is_array)
    arrays_tree = tree_map(
        lambda x: (
            type(x),
            "True" if eqx.is_array(x) else ("False" if x is not None else "None"),
        ),
        arrays,
    )
    jax.debug.print("{name} arrays_tree = {out}", name=name, out=arrays_tree)

    static_tree = tree_map(
        lambda x: (
            type(x),
            "True" if eqx.is_array(x) else ("False" if x is not None else "None"),
        ),
        static,
    )
    jax.debug.print("{name} static_tree = {out}", name=name, out=static_tree)



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Core classes and functions for thermochemical data"""

import equinox as eqx
import jax.numpy as jnp
from jax import Array
from jax.typing import ArrayLike
from molmass import Formula
from xmmutablemap import ImmutableMap

from atmodeller.utilities import unit_conversion


class CondensateActivity(eqx.Module):
    """Activity of a stable condensate"""

    activity: ArrayLike = 1.0

    @eqx.filter_jit
    def log_activity(
        self,
        temperature: ArrayLike,
        pressure: ArrayLike,
    ) -> ArrayLike:
        del temperature
        del pressure

        return jnp.log(self.activity)


class ThermoCoefficients(eqx.Module):
    """Coefficients for thermochemical data

    Coefficients are available at https://ntrs.nasa.gov/citations/20020085330

    Args:
        b1: Enthalpy constant(s) of integration
        b2: Entropy constant(s) of integration
        cp_coeffs: Heat capacity coefficients
        T_min: Minimum temperature(s) in the range
        T_max: Maximum temperature(s) in the range

    Attributes:
        b1: Enthalpy constant(s) of integration
        b2: Entropy constant(s) of integration
        cp_coeffs: Heat capacity coefficients
        T_min: Minimum temperature(s) in the range
        T_max: Maximum temperature(s) in the range
    """

    b1: tuple[float, ...]
    b2: tuple[float, ...]
    cp_coeffs: tuple[tuple[float, float, float, float, float, float, float], ...]
    T_min: tuple[float, ...]
    T_max: tuple[float, ...]

    @eqx.filter_jit
    def _cp_over_R(self, cp_coefficients: ArrayLike, temperature: ArrayLike) -> Array:
        """Heat capacity relative to the gas constant (R)

        Args:
            cp_coefficients: Heat capacity coefficients as an array
            temperature: Temperature

        Returns:
            Heat capacity (J/K/mol) relative to R
        """
        temperature_terms: Array = jnp.stack(
            [
                jnp.power(temperature, -2),
                jnp.power(temperature, -1),
                jnp.ones_like(temperature),
                temperature,
                jnp.power(temperature, 2),
                jnp.power(temperature, 3),
                jnp.power(temperature, 4),
            ]
        )

        heat_capacity: Array = jnp.dot(cp_coefficients, temperature_terms)
        # jax.debug.print("heat_capacity = {out}", out=heat_capacity)

        return heat_capacity

    @eqx.filter_jit
    def _S_over_R(
        self, cp_coefficients: ArrayLike, b2: ArrayLike, temperature: ArrayLike
    ) -> Array:
        """Entropy relative to the gas constant (R)

        Args:
            cp_coefficients: Heat capacity coefficients as an array
            b2: Entropy integration constant
            temperature: Temperature

        Returns:
            Entropy (J/K/mol) relative to R
        """
        temperature_terms: Array = jnp.stack(
            [
                -jnp.power(temperature, -2) / 2,
                -jnp.power(temperature, -1),
                jnp.log(temperature),
                temperature,
                jnp.power(temperature, 2) / 2,
                jnp.power(temperature, 3) / 3,
                jnp.power(temperature, 4) / 4,
            ]
        )

        entropy: Array = jnp.dot(cp_coefficients, temperature_terms) + b2
        # jax.debug.print("entropy = {out}", out=entropy)

        return entropy

    @eqx.filter_jit
    def _H_over_RT(
        self, cp_coefficients: ArrayLike, b1: ArrayLike, temperature: ArrayLike
    ) -> Array:
        """Enthalpy relative to RT

        Args:
            cp_coefficients: Heat capacity coefficients as an array
            b1: Enthalpy integration constant
            temperature: Temperature

        Returns:
            Enthalpy (J/mol) relative to RT
        """
        temperature_terms: Array = jnp.stack(
            [
                -jnp.power(temperature, -2),
                jnp.log(temperature) / temperature,
                jnp.ones_like(temperature),
                temperature / 2,
                jnp.power(temperature, 2) / 3,
                jnp.power(temperature, 3) / 4,
                jnp.power(temperature, 4) / 5,
            ]
        )

        enthalpy: Array = jnp.dot(cp_coefficients, temperature_terms) + b1 / temperature
        # jax.debug.print("enthalpy = {out}", out=enthalpy)

        return enthalpy

    @eqx.filter_jit
    def _G_over_RT(
        self, cp_coefficients: ArrayLike, b1: ArrayLike, b2: ArrayLike, temperature: ArrayLike
    ) -> Array:
        """Gibbs energy relative to RT

        Args:
            cp_coefficients: Heat capacity coefficients as an array
            b1: Enthalpy integration constant
            b2: Entropy integration constant
            temperature: Temperature

        Returns:
            Gibbs energy relative to RT
        """
        enthalpy: Array = self._H_over_RT(cp_coefficients, b1, temperature)
        entropy: Array = self._S_over_R(cp_coefficients, b2, temperature)
        # No temperature multiplication is correct since the return is Gibbs energy relative to RT
        gibbs: Array = enthalpy - entropy

        return gibbs

    @eqx.filter_jit
    def get_gibbs_over_RT(self, temperature: ArrayLike) -> Array:
        """Gets Gibbs energy over RT

        This is calculated using data from the appropriate temperature range.

        Args:
            temperature: Temperature

        Returns:
            Gibbs energy over RT
        """
        # This assumes the temperature is within one of the ranges and will produce unexpected
        # output if the temperature is outside the ranges
        T_min_array: Array = jnp.asarray(self.T_min)
        T_max_array: Array = jnp.asarray(self.T_max)
        # Temperature must be a float array since JAX cannot raise integers to negative powers.
        temperature = jnp.asarray(temperature, dtype=jnp.float_)
        bool_mask: Array = (T_min_array <= temperature) & (temperature <= T_max_array)
        index: Array = jnp.argmax(bool_mask)
        # jax.debug.print("index = {out}", out=index)
        cp_coeffs_for_index: Array = jnp.take(jnp.array(self.cp_coeffs), index, axis=0)
        # jax.debug.print("cp_coeffs_for_index = {out}", out=cp_coeffs_for_index)
        b1_for_index: Array = jnp.take(jnp.array(self.b1), index)
        # jax.debug.print("b1_for_index = {out}", out=b1_for_index)
        b2_for_index: Array = jnp.take(jnp.array(self.b2), index)
        # jax.debug.print("b2_for_index = {out}", out=b2_for_index)
        gibbs_for_index: Array = self._G_over_RT(
            cp_coeffs_for_index, b1_for_index, b2_for_index, temperature
        )

        return gibbs_for_index


class SpeciesData(eqx.Module):
    """Species data

    Args:
        formula: Formula
        phase: Phase
        thermodata: Thermodynamic data
    """

    formula: str
    """Formula"""
    phase: str
    """Phase"""
    thermodata: ThermoCoefficients
    """Thermodynamic data"""
    composition: ImmutableMap[str, tuple[int, float, float]] = eqx.field(init=False)
    """Composition"""
    hill_formula: str = eqx.field(init=False)
    """Hill formula"""
    molar_mass: float = eqx.field(init=False)
    """Molar mass"""

    def __post_init__(self):
        mformula: Formula = Formula(self.formula)
        self.composition = ImmutableMap(mformula.composition().asdict())
        self.hill_formula = mformula.formula
        self.molar_mass = mformula.mass * unit_conversion.g_to_kg

    @property
    def elements(self) -> tuple[str, ...]:
        """Elements"""
        return tuple(self.composition.keys())

    @property
    def name(self) -> str:
        """Unique name by combining Hill notation and phase"""
        return f"{self.hill_formula}_{self.phase}"

    def get_gibbs_over_RT(self, temperature: ArrayLike) -> Array:
        """Gets Gibbs energy over RT

        This is calculated using data from the appropriate temperature range.

        Args:
            temperature: Temperature

        Returns:
            Gibbs energy over RT
        """
        return self.thermodata.get_gibbs_over_RT(temperature)


class CriticalData(eqx.Module):
    """Critical temperature and pressure of a gas species

    Args:
        temperature: Critical temperature in K
        pressure: Critical pressure in bar
    """

    temperature: float = 1
    """Critical temperature in K"""
    pressure: float = 1
    """Critical pressure in bar"""



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Thermodata package level variables"""

# Expose public API
from atmodeller.thermodata._core import (  # noqa: F401
    CondensateActivity,
    CriticalData,
    SpeciesData,
    ThermoCoefficients,
)
from atmodeller.thermodata._redox_buffers import IronWustiteBuffer  # noqa: F401
from atmodeller.thermodata.library import (  # noqa: F401
    get_thermodata,
    select_critical_data,
    select_thermodata,
)



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""EOS package"""

import importlib.resources
from importlib.abc import Traversable

DATA_DIRECTORY: Traversable = importlib.resources.files(f"{__package__}.data")
"""Data directory, which is the same as the package directory"""
# ABSOLUTE_TOLERANCE is less than RELATIVE_TOLERANCE because typical volumes are around 1e-6
ABSOLUTE_TOLERANCE: float = 1.0e-12
r"""Absolute tolerance when solving for the volume in :math:`\mathrm{m}^3\ \mathrm{mol}^{-1}`"""
RELATIVE_TOLERANCE: float = 1.0e-6
r"""Relative tolerance when solving for the volume in :math:`\mathrm{m}^3\ \mathrm{mol}^{-1}`"""
THROW: bool = False
"""Whether to throw errors. Change to True for debugging purposes."""

# Expose the public API
from atmodeller.eos._aggregators import CombinedRealGas  # noqa: E402, F401
from atmodeller.eos.core import IdealGas, RealGas  # noqa: E402, F401
from atmodeller.eos.library import get_eos_models  # noqa: E402, F401



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
from molmass import Formula
from scipy import constants

AVOGADRO: float = constants.Avogadro
"""Avogadro constant in 1/mol"""
GAS_CONSTANT: float = constants.gas_constant
"""Gas constant in J/K/mol"""
GAS_CONSTANT_BAR: float = GAS_CONSTANT * 1.0e-5
"""Gas constant in m^3 bar/K/mol"""
GRAVITATIONAL_CONSTANT: float = constants.gravitational_constant
"""Gravitational constant in m^3/kg/s^2"""
ATMOSPHERE: float = constants.atmosphere / constants.bar
"""Atmospheres in 1 bar"""
BOLTZMANN_CONSTANT: float = constants.Boltzmann
"""Boltzmann constant in J/K"""
BOLTZMANN_CONSTANT_BAR: float = BOLTZMANN_CONSTANT * 1e-5
"""Boltzmann constant in bar m^3/K"""
OCEAN_MOLES: float = 7.68894973907177e22
"""Moles of H2 or H2O in one present-day Earth ocean"""
OCEAN_MASS_H2: float = OCEAN_MOLES * Formula("H2").mass
"""Mass of H2 in one present-day Earth ocean in grams"""
OCEAN_MASS_H2O: float = OCEAN_MOLES * Formula("H2O").mass
"""Mass of H2O in one present-day Earth ocean in grams"""



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Solubility package"""

# Expose the public API
from atmodeller.solubility.core import (  # noqa: E402, F401
    NoSolubility,
    Solubility,
    SolubilityPowerLaw,
    SolubilityPowerLawLog10,
)
from atmodeller.solubility.library import get_solubility_models  # noqa: E402, F401



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Output

This uses existing functions as much as possible to calculate desired output quantities, where some
must be vmapped to compute the output.
"""

from __future__ import annotations

import logging
import pickle
from pathlib import Path
from typing import Any, Callable

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
import numpy.typing as npt
import pandas as pd
from jax import Array
from jax.tree_util import tree_map
from jax.typing import ArrayLike
from molmass import Formula
from openpyxl.styles import PatternFill
from scipy.constants import mega

from atmodeller.constants import AVOGADRO
from atmodeller.containers import (
    FixedParameters,
    Planet,
    Species,
    SpeciesCollection,
    TracedParameters,
)
from atmodeller.engine import (
    get_atmosphere_log_molar_mass,
    get_atmosphere_log_volume,
    get_element_density,
    get_element_density_in_melt,
    get_log_activity,
    get_log_number_density_from_log_pressure,
    get_pressure_from_log_number_density,
    get_species_density_in_melt,
    get_species_ppmw_in_melt,
    get_total_pressure,
    objective_function,
)
from atmodeller.interfaces import RedoxBufferProtocol
from atmodeller.thermodata import IronWustiteBuffer
from atmodeller.utilities import unit_conversion, vmap_axes_spec

logger: logging.Logger = logging.getLogger(__name__)


class Output:
    """Output

    Args:
        species: Species
        solution: Array output from solve
        solver_status: Solver status
        solver_steps: Number of solver steps
        fixed_parameters: Fixed parameters
        traced_parameters: Traced parameters
    """

    def __init__(
        self,
        species: SpeciesCollection,
        solution: Array,
        solver_status: Array,
        solver_steps: Array,
        fixed_parameters: FixedParameters,
        traced_parameters: TracedParameters,
    ):
        logger.debug("Creating Output")
        self._species: SpeciesCollection = species
        self._solution: Array = solution
        self._solver_status: Array = solver_status
        self._solver_steps: Array = solver_steps
        self._fixed_parameters: FixedParameters = fixed_parameters
        self._traced_parameters: TracedParameters = traced_parameters

        # Calculate the index at which to split the array
        split_index: int = self._solution.shape[1] - self._species.number_of_stability()
        log_number_density, log_stability = jnp.split(self._solution, [split_index], axis=1)
        # Number of entries in log_number_density is always the total number of species
        self._log_number_density: Array = log_number_density
        # Number of entries in log_stability could be between 0 and the total number of species
        self._log_stability: Array = log_stability
        # Caching output to avoid recomputation
        self._cached_dict: dict[str, dict[str, Array]] | None = None
        self._cached_dataframes: dict[str, pd.DataFrame] | None = None

    @property
    def condensed_species_indices(self) -> Array:
        """Condensed species indices"""
        return self._species.get_condensed_species_indices()

    @property
    def gas_species_indices(self) -> Array:
        """Gas species indices"""
        return self._fixed_parameters.gas_species_indices

    @property
    def log_number_density(self) -> Array:
        """Log number density"""
        return self._log_number_density

    @property
    def log_stability(self) -> Array:
        """Log stability of relevant species"""
        return self._log_stability

    @property
    def molar_mass(self) -> Array:
        """Molar mass of all species"""
        return self._fixed_parameters.molar_masses

    @property
    def number_solutions(self) -> int:
        """Number of solutions"""
        return self.log_number_density.shape[0]

    @property
    def planet(self) -> Planet:
        """Planet"""
        return self._traced_parameters.planet

    @property
    def stability_species_mask(self) -> Array:
        """Stability species mask"""
        return self._species.get_stability_species_mask()

    @property
    def temperature(self) -> Array:
        """Temperature"""
        return jnp.asarray(self.planet.temperature)

    @property
    def temperature_vmap(self) -> int | None:
        """Axis for temperature vmap"""
        return vmap_axes_spec(self._traced_parameters.planet).temperature

    @property
    def traced_parameters_vmap(self) -> TracedParameters:
        """Axis for traced parameters vmap"""
        return vmap_axes_spec(self._traced_parameters)

    def activity(self) -> Array:
        """Gets the activity of all species

        Returns:
            Activity of all species
        """
        return jnp.exp(self.log_activity())

    def activity_without_stability(self) -> Array:
        """Gets the activity without stability of all species

        Returns:
            Activity without stability of all species
        """
        return jnp.exp(self.log_activity_without_stability())

    def asdict(self) -> dict[str, dict[str, Array]]:
        """All output in a dictionary, with caching.

        TODO: This function is slow and is an obvious candidate for speeding up.

        Returns:
            Dictionary of all output
        """
        if self._cached_dict is not None:
            logger.info("Returning cached asdict output")
            return self._cached_dict  # Return cached result

        logger.info("Computing asdict output")
        fugacity_matrix: Array = jnp.array(self._fixed_parameters.fugacity_matrix)
        formula_matrix_constraints: Array = jnp.array(
            self._fixed_parameters.formula_matrix_constraints
        )

        out: dict[str, dict[str, Array]] = {}
        out |= self.condensed_species_asdict()
        out |= self.gas_species_asdict()
        out |= self.elements_asdict()

        out["planet"] = expand_dict(self.planet.asdict(), self.number_solutions)
        out["atmosphere"] = self.atmosphere_asdict()
        # temperature and pressure have already been expanded to the number of solutions
        temperature: Array = out["planet"]["surface_temperature"]
        pressure: Array = out["atmosphere"]["pressure"]
        # Convenient to also attach temperature to the atmosphere output
        out["atmosphere"]["temperature"] = temperature
        out["raw_solution"] = self.raw_solution_asdict()

        out["constraints"] = {}
        if formula_matrix_constraints.size > 0:
            out["constraints"] |= expand_dict(
                self._traced_parameters.mass_constraints.asdict(), self.number_solutions
            )
        if fugacity_matrix.size > 0:
            out["constraints"] |= self._traced_parameters.fugacity_constraints.asdict(
                temperature, pressure
            )
        out["residual"] = self.residual_asdict()  # type: ignore since uses int for keys

        if "O2_g" in out:
            logger.debug("Found O2_g so back-computing log10 shift for fO2")
            log10_fugacity: Array = jnp.log10(out["O2_g"]["fugacity"])
            temperature: Array = out["planet"]["surface_temperature"]
            pressure: Array = out["atmosphere"]["pressure"]
            buffer: RedoxBufferProtocol = IronWustiteBuffer()
            # Shift at 1 bar
            buffer_at_one_bar: ArrayLike = buffer.log10_fugacity(temperature, 1.0)
            log10_shift_at_one_bar: Array = log10_fugacity - buffer_at_one_bar
            logger.debug("log10_shift_at_1bar = %s", log10_shift_at_one_bar)
            out["O2_g"]["log10dIW_1_bar"] = log10_shift_at_one_bar
            # Shift at actual pressure
            buffer_at_P: ArrayLike = buffer.log10_fugacity(temperature, pressure)
            log10_shift_at_P: Array = log10_fugacity - buffer_at_P
            logger.debug("log10_shift_at_P = %s", log10_shift_at_P)
            out["O2_g"]["log10dIW_P"] = log10_shift_at_P

        out["solver"] = {"status": self._solver_status, "steps": self._solver_steps}

        # Convert all arrays in the dictionary to numpy arrays. Using the same functions that the
        # engine uses makes sense to avoid duplication and ensure consistency, but for the output
        # we can convert all arrays to numpy. Then the output will only consistent of dictionaries
        # and numpy arrays, which might simplify the pickling/unpickling process for end-users who
        # only care about the results and do not necessary want to use JAX types to process
        # the output. To this point the arrays are of type
        # <class 'jaxlib.xla_extension.ArrayImpl'>.
        # TODO: could maybe use numpy method?
        def convert_to_numpy(d) -> None:
            for key, value in d.items():
                if isinstance(value, dict):
                    convert_to_numpy(value)
                else:
                    logger.debug("Array type before conversion = %s", type(d[key]))
                    d[key] = np.array(value)
                    logger.debug("Array type after conversion = %s", type(d[key]))

        logger.debug("Convert all arrays to numpy")
        convert_to_numpy(out)

        self._cached_dict = out  # Cache result

        return out

    def atmosphere_asdict(self) -> dict[str, Array]:
        """Gets the atmosphere properties

        Returns:
            Atmosphere properties
        """
        out: dict[str, Array] = {}

        log_number_density_from_log_pressure_func: Callable = eqx.filter_vmap(
            get_log_number_density_from_log_pressure, in_axes=(0, self.temperature_vmap)
        )
        log_number_density = log_number_density_from_log_pressure_func(
            jnp.log(self.total_pressure()), self.temperature
        )
        # Must be 2-D to align arrays for computing number-density-related quantities
        number_density: Array = jnp.exp(log_number_density)[:, jnp.newaxis]
        molar_mass: Array = self.atmosphere_molar_mass()[:, jnp.newaxis]
        out: dict[str, Array] = self._get_number_density_output(
            number_density, molar_mass, "species_"
        )
        # Species mass is simply mass, so rename for clarity
        out["mass"] = out.pop("species_mass")

        out["molar_mass"] = molar_mass
        # Ensure all arrays are 1-D, which is required for creating dataframes
        out = {key: value.ravel() for key, value in out.items()}

        out["pressure"] = self.total_pressure()
        out["volume"] = self.atmosphere_volume()
        out["element_number_density"] = jnp.sum(self.element_density_gas(), axis=1)
        out["element_number"] = out["element_number_density"] * out["volume"]
        out["element_moles"] = out["element_number"] / AVOGADRO

        return out

    def atmosphere_log_molar_mass(self) -> Array:
        """Gets log molar mass of the atmosphere

        Returns:
            Log molar mass of the atmosphere
        """
        atmosphere_log_molar_mass_func: Callable = eqx.filter_vmap(
            get_atmosphere_log_molar_mass, in_axes=(None, 0)
        )
        atmosphere_log_molar_mass: Array = atmosphere_log_molar_mass_func(
            self._fixed_parameters, self.log_number_density
        )

        return atmosphere_log_molar_mass

    def atmosphere_molar_mass(self) -> Array:
        """Gets the molar mass of the atmosphere

        Returns:
            Molar mass of the atmosphere
        """
        return jnp.exp(self.atmosphere_log_molar_mass())

    def atmosphere_log_volume(self) -> Array:
        """Gets the log volume of the atmosphere

        Returns:
            Log volume of the atmosphere
        """
        atmosphere_log_volume_func: Callable = eqx.filter_vmap(
            get_atmosphere_log_volume,
            in_axes=(
                None,
                0,
                vmap_axes_spec(self._traced_parameters.planet),
            ),
        )
        atmosphere_log_volume: Array = atmosphere_log_volume_func(
            self._fixed_parameters,
            self.log_number_density,
            self.planet,
        )

        return atmosphere_log_volume

    def atmosphere_volume(self) -> Array:
        """Gets the volume of the atmosphere

        Returns:
            Volume of the atmosphere
        """
        return jnp.exp(self.atmosphere_log_volume())

    def total_pressure(self) -> Array:
        """Gets total pressure

        Returns:
            Total pressure
        """
        total_pressure_func: Callable = eqx.filter_vmap(
            get_total_pressure, in_axes=(None, 0, self.temperature_vmap)
        )
        total_pressure: Array = total_pressure_func(
            self._fixed_parameters, self.log_number_density, self.temperature
        )

        return total_pressure

    def condensed_species_asdict(self) -> dict[str, dict[str, Array]]:
        """Gets the condensed species output as a dictionary

        Returns:
            Condensed species output as a dictionary
        """
        molar_mass: Array = self.species_molar_mass_expanded()
        number_density_condensed: Array = jnp.take(
            self.number_density(), self.condensed_species_indices, axis=1
        )
        activity_condensed: Array = jnp.take(
            self.activity(), self.condensed_species_indices, axis=1
        )
        molar_mass_condensed: Array = jnp.take(molar_mass, self.condensed_species_indices, axis=1)
        condensed_species: tuple[Species, ...] = tuple(
            self._species[ii] for ii in self.condensed_species_indices
        )
        out: dict[str, Array] = self._get_number_density_output(
            number_density_condensed, molar_mass_condensed, "total_"
        )
        out["molar_mass"] = molar_mass_condensed
        out["activity"] = activity_condensed

        split_dict: list[dict[str, Array]] = split_dict_by_columns(out)
        species_out: dict[str, dict[str, Array]] = {
            species.name: split_dict[ii] for ii, species in enumerate(condensed_species)
        }

        return species_out

    def elements_asdict(self) -> dict[str, dict[str, Array]]:
        """Gets the element properties as a dictionary

        Returns:
            Element outputs as a dictionary
        """
        molar_mass: Array = self.element_molar_mass_expanded()
        atmosphere: Array = self.element_density_gas()
        condensed: Array = self.element_density_condensed()
        dissolved: Array = self.element_density_dissolved()
        total: Array = atmosphere + condensed + dissolved

        out: dict[str, Array] = self._get_number_density_output(
            atmosphere, molar_mass, "atmosphere_"
        )
        out |= self._get_number_density_output(condensed, molar_mass, "condensed_")
        out |= self._get_number_density_output(dissolved, molar_mass, "dissolved_")
        out |= self._get_number_density_output(total, molar_mass, "total_")

        out["molar_mass"] = molar_mass
        out["degree_of_condensation"] = out["condensed_number"] / out["total_number"]
        out["volume_mixing_ratio"] = out["atmosphere_number"] / jnp.sum(
            out["atmosphere_number"], axis=1, keepdims=True
        )
        out["atmosphere_ppm"] = out["volume_mixing_ratio"] * mega
        out["atmosphere_ppmw"] = (
            out["atmosphere_mass"] / jnp.sum(out["atmosphere_mass"], axis=1, keepdims=True) * mega
        )

        unique_elements: tuple[str, ...] = self._species.get_unique_elements_in_species()
        if "H" in unique_elements:
            index: int = unique_elements.index("H")
            H_total_moles: Array = out["total_moles"][:, index]
            out["logarithmic_abundance"] = (
                jnp.log10(out["total_moles"] / H_total_moles[:, jnp.newaxis]) + 12
            )

        logger.debug("out = %s", out)

        split_dict: list[dict[str, Array]] = split_dict_by_columns(out)
        logger.debug("split_dict = %s", split_dict)

        elements_out: dict[str, dict[str, Array]] = {
            f"element_{element}": split_dict[ii] for ii, element in enumerate(unique_elements)
        }
        logger.debug("elements_out = %s", elements_out)

        return elements_out

    def element_density_condensed(self) -> Array:
        """Gets the number density of elements in the condensed phase

        Unlike for the objective function, we want the number density of all elements, regardless
        of whether they were used to impose a mass constraint on the system.

        Returns:
            Number density of elements in the condensed phase
        """
        element_density_func: Callable = eqx.filter_vmap(get_element_density, in_axes=(None, 0))
        element_density: Array = element_density_func(
            self.formula_matrix_condensed(), self.log_number_density
        )

        return element_density

    def element_density_dissolved(self) -> Array:
        """Gets the number density of elements dissolved in melt due to species solubility

        Unlike for the objective function, we want the number density of all elements, regardless
        of whether they were used to impose a mass constraint on the system.

        Returns:
            Number density of elements dissolved in melt due to species solubility
        """
        element_density_dissolved_func: Callable = eqx.filter_vmap(
            get_element_density_in_melt,
            in_axes=(self.traced_parameters_vmap, None, None, 0, 0, 0),
        )
        element_density_dissolved: Array = element_density_dissolved_func(
            self._traced_parameters,
            self._fixed_parameters,
            jnp.array(self._fixed_parameters.formula_matrix),
            self.log_number_density,
            self.log_activity(),
            self.atmosphere_log_volume(),
        )

        return element_density_dissolved

    def element_density_gas(self) -> Array:
        """Gets the number density of elements in the gas phase

        Unlike for the objective function, we want the number density of all elements, regardless
        of whether they were used to impose a mass constraint on the system.

        Returns:
            Number density of elements in the gas phase
        """
        element_density_func: Callable = eqx.filter_vmap(get_element_density, in_axes=(None, 0))
        element_density: Array = element_density_func(
            self.formula_matrix_gas(), self.log_number_density
        )

        return element_density

    def element_molar_mass_expanded(self) -> Array:
        """Gets molar mass of elements

        Returns:
            Molar mass of elements
        """
        unique_elements: tuple[str, ...] = self._species.get_unique_elements_in_species()
        molar_mass: Array = jnp.array([Formula(element).mass for element in unique_elements])
        molar_mass = unit_conversion.g_to_kg * molar_mass

        return jnp.tile(molar_mass, (self.number_solutions, 1))

    def _get_modified_formula_matrix(self, indices: Array) -> Array:
        """Gets a modified formula matrix with columns not in indices set to zero.

        Args:
            indices: Indices of columns to keep, where others are set to zero.

        Returns:
            Modified formula matrix
        """
        formula_matrix: Array = jnp.array(self._fixed_parameters.formula_matrix)
        mask: Array = jnp.zeros_like(formula_matrix, dtype=bool)
        mask = mask.at[:, indices].set(True)
        logger.debug("modified_formula_matrix = %s", mask)
        modified_formula_matrix: Array = formula_matrix * mask

        return modified_formula_matrix

    def formula_matrix_condensed(self) -> Array:
        """Formula matrix for condensed species

        Only columns in the formula matrix referring to condensed species have non-zero values.

        Returns:
            Formula matrix for condensed species
        """
        return self._get_modified_formula_matrix(self.condensed_species_indices)

    def formula_matrix_gas(self) -> Array:
        """Formula matrix for gas species

        Only columns in the formula matrix referring to gas species have non-zero values.

        Returns:
            Formula matrix for gas species
        """
        return self._get_modified_formula_matrix(self.gas_species_indices)

    def _get_number_density_output(
        self, number_density: Array, molar_mass_expanded: Array, prefix: str = ""
    ) -> dict[str, Array]:
        """Gets the outputs associated with a number density

        Args:
            number_density: Number density
            molar_mass_expanded: Molar mass associated with the number density
            prefix: Key prefix for the output. Defaults to an empty string.

        Returns
            Dictionary of output quantities
        """
        atmosphere_volume: Array = self.atmosphere_volume()
        # Volume must be a column vector because it multiples all elements in the row
        number: Array = number_density * atmosphere_volume[:, jnp.newaxis]
        moles: Array = number / AVOGADRO
        mass: Array = moles * molar_mass_expanded

        out: dict[str, Array] = {}
        out[f"{prefix}number_density"] = number_density
        out[f"{prefix}number"] = number
        out[f"{prefix}moles"] = moles
        out[f"{prefix}mass"] = mass

        return out

    def gas_species_asdict(self) -> dict[str, dict[str, Array]]:
        """Gets the gas species output as a dictionary

        Returns:
            Gas species output as a dictionary
        """
        molar_mass: Array = self.species_molar_mass_expanded()
        number_density_gas: Array = jnp.take(
            self.number_density(), self.gas_species_indices, axis=1
        )
        dissolved_gas: Array = jnp.take(
            self.species_density_in_melt(), self.gas_species_indices, axis=1
        )
        total_gas: Array = number_density_gas + dissolved_gas
        activity_gas: Array = jnp.take(self.activity(), self.gas_species_indices, axis=1)
        pressure_gas: Array = jnp.take(self.pressure(), self.gas_species_indices, axis=1)
        molar_mass_gas: Array = jnp.take(molar_mass, self.gas_species_indices, axis=1)
        gas_species: tuple[Species, ...] = tuple(
            self._species[ii] for ii in self.gas_species_indices
        )

        out: dict[str, Array] = {}
        out |= self._get_number_density_output(number_density_gas, molar_mass_gas, "atmosphere_")
        out |= self._get_number_density_output(dissolved_gas, molar_mass_gas, "dissolved_")
        out |= self._get_number_density_output(total_gas, molar_mass_gas, "total_")
        out["molar_mass"] = molar_mass
        out["volume_mixing_ratio"] = out["atmosphere_number"] / jnp.sum(
            out["atmosphere_number"], axis=1, keepdims=True
        )
        out["atmosphere_ppm"] = out["volume_mixing_ratio"] * mega
        out["atmosphere_ppmw"] = (
            out["atmosphere_mass"] / jnp.sum(out["atmosphere_mass"], axis=1, keepdims=True) * mega
        )
        out["pressure"] = pressure_gas
        out["fugacity"] = activity_gas
        out["fugacity_coefficient"] = activity_gas / pressure_gas
        out["dissolved_ppmw"] = self.species_ppmw_in_melt()

        split_dict: list[dict[str, Array]] = split_dict_by_columns(out)
        species_out: dict[str, dict[str, Array]] = {
            species.name: split_dict[ii] for ii, species in enumerate(gas_species)
        }

        return species_out

    def log_activity(self) -> Array:
        """Gets log activity of all species.

        This is usually what the user wants when referring to activity because it includes a
        consideration of species stability

        Returns:
            Log activity of all species
        """
        log_activity_without_stability: Array = self.log_activity_without_stability()
        # We select the relevant log_activity at the end of this method depending which species
        # also have stability, so its OK to build a padded array here to keep the size the same.
        log_stability_padded: Array = jnp.zeros_like(log_activity_without_stability)
        log_stability_padded = log_stability_padded.at[
            :, self._species.get_stability_species_indices()
        ].set(self.log_stability)
        # Note that the below array contains incorrect entries for species without stability since
        # exp(0)=1. But these are filtered out further below.
        log_activity_with_stability: Array = log_activity_without_stability - jnp.exp(
            log_stability_padded
        )

        # Now select the appropriate activity for each species, depending if stability is relevant.
        condition_broadcasted = jnp.broadcast_to(
            self.stability_species_mask, log_activity_without_stability.shape
        )
        # logger.debug("condition_broadcasted = %s", condition_broadcasted)

        log_activity: Array = jnp.where(
            condition_broadcasted,
            log_activity_with_stability,
            log_activity_without_stability,
        )

        return log_activity

    def log_activity_without_stability(self) -> Array:
        """Gets log activity without stability of all species

        Args:
            Log activity without stability of all species
        """
        log_activity_func: Callable = eqx.filter_vmap(
            get_log_activity,
            in_axes=(vmap_axes_spec(self._traced_parameters), None, 0),
        )
        log_activity: Array = log_activity_func(
            self._traced_parameters, self._fixed_parameters, self.log_number_density
        )

        return log_activity

    def number_density(self) -> Array:
        r"""Gets number density of all species

        Returns:
            Number density in :math:`\mathrm{molecules}\, \mathrm{m}^{-3}`
        """
        return jnp.exp(self.log_number_density)

    def species_molar_mass_expanded(self) -> Array:
        r"""Gets molar mass of all species in an expanded array.

        Returns:
            Molar mass of all species in an expanded array.
        """
        return jnp.tile(self.molar_mass, (self.number_solutions, 1))

    def pressure(self) -> Array:
        """Gets pressure of species in bar

        This will compute pressure of all species, including condensates, for simplicity.

        Returns:
            Pressure of species in bar
        """
        pressure_func: Callable = eqx.filter_vmap(
            get_pressure_from_log_number_density, in_axes=(0, self.temperature_vmap)
        )
        pressure: Array = pressure_func(self.log_number_density, self.temperature)

        return pressure

    def quick_look(self) -> dict[str, ArrayLike]:
        """Quick look at the solution

        Provides a quick first glance at the output with convenient units and to ease comparison
        with test or benchmark data.

        Returns:
            Dictionary of the solution
        """
        out: dict[str, ArrayLike] = {}

        for nn, species_ in enumerate(self._species):
            pressure: Array = self.pressure()[:, nn]
            activity: Array = self.activity()[:, nn]
            out[species_.name] = pressure
            out[f"{species_.name}_activity"] = activity

        return collapse_single_entry_values(out)

    def raw_solution_asdict(self) -> dict[str, Array]:
        """Gets the raw solution

        Returns:
            Dictionary of the raw solution
        """
        raw_solution: dict[str, Array] = {}

        species_names: tuple[str, ...] = self._species.get_species_names()

        for ii, species_name in enumerate(species_names):
            raw_solution[species_name] = self.log_number_density[:, ii]

        for ii, index in enumerate(self._species.get_stability_species_indices()):
            species_name = species_names[index]
            raw_solution[f"{species_name}_stability"] = self.log_stability[:, ii]

        return raw_solution

    def residual_asdict(self) -> dict[int, Array]:
        """Gets the residual

        Returns:
            Dictionary of the residual
        """
        residual_func: Callable = eqx.filter_vmap(
            objective_function,
            in_axes=(
                0,
                {
                    "traced_parameters": self.traced_parameters_vmap,
                    "fixed_parameters": None,
                },
            ),
        )
        residual: Array = residual_func(
            self._solution,
            {
                "traced_parameters": self._traced_parameters,
                "fixed_parameters": self._fixed_parameters,
            },
        )

        out: dict[int, Array] = {}
        for ii in range(residual.shape[1]):
            out[ii] = residual[:, ii]

        return out

    def species_density_in_melt(self) -> Array:
        """Gets species number density in the melt

        Returns:
            Species number density in the melt
        """
        species_density_in_melt_func: Callable = eqx.filter_vmap(
            get_species_density_in_melt,
            in_axes=(self.traced_parameters_vmap, None, 0, 0, 0),
        )
        species_density_in_melt: Array = species_density_in_melt_func(
            self._traced_parameters,
            self._fixed_parameters,
            self.log_number_density,
            self.log_activity(),
            self.atmosphere_log_volume(),
        )

        return species_density_in_melt

    def species_ppmw_in_melt(self) -> Array:
        """Gets species ppmw in the melt

        Return:
            Species ppmw in the melt
        """
        species_ppmw_in_melt_func: Callable = eqx.filter_vmap(
            get_species_ppmw_in_melt, in_axes=(self.traced_parameters_vmap, None, 0, 0)
        )
        species_ppmw_in_melt: Array = species_ppmw_in_melt_func(
            self._traced_parameters,
            self._fixed_parameters,
            self.log_number_density,
            self.log_activity(),
        )

        return species_ppmw_in_melt

    def stability(self) -> Array:
        """Gets stability of relevant species

        Returns:
            Stability of relevant species
        """
        return jnp.exp(self.log_stability)

    def _drop_unsuccessful_solves(
        self, dataframes: dict[str, pd.DataFrame]
    ) -> dict[str, pd.DataFrame]:
        """Drops unsuccessful solves

        Args:
            dataframes: Dataframes from which to drop unsuccessful models

        Returns:
            Dictionary of dataframes without unsuccessful models
        """
        return {key: df.loc[np.array(self._solver_status)] for key, df in dataframes.items()}

    def to_dataframes(self, drop_unsuccessful: bool = False) -> dict[str, pd.DataFrame]:
        """Gets the output in a dictionary of dataframes.

        Args:
            drop_unsuccessful: Drop models that did not solve. Defaults to False.

        Returns:
            Output in a dictionary of dataframes
        """
        if self._cached_dataframes is not None:
            logger.debug("Returning cached to_dataframes output")
            dataframes: dict[str, pd.DataFrame] = self._cached_dataframes  # Return cached result
        else:
            logger.info("Computing to_dataframes output")
            dataframes = nested_dict_to_dataframes(self.asdict())
            self._cached_dataframes = dataframes
            logger.debug("to_dataframes = %s", self._cached_dataframes)

        if drop_unsuccessful:
            logger.info("Dropping models that did not solve")
            dataframes: dict[str, pd.DataFrame] = self._drop_unsuccessful_solves(dataframes)

        return dataframes

    def to_excel(
        self, file_prefix: Path | str = "new_atmodeller_out", drop_unsuccessful: bool = False
    ) -> None:
        """Writes the output to an Excel file.

        Args:
            file_prefix: Prefix of the output file. Defaults to new_atmodeller_out.
            drop_unsuccessful: Drop models that did not solve. Defaults to False.
        """
        logger.info("Writing output to excel")
        out: dict[str, pd.DataFrame] = self.to_dataframes(drop_unsuccessful)
        output_file: Path = Path(f"{file_prefix}.xlsx")

        # Convenient to highlight rows where the solver failed to find a solution for follow-up
        # analysis
        # Define a fill color for highlighting rows (e.g., yellow)
        highlight_fill = PatternFill(start_color="FFFF00", end_color="FFFF00", fill_type="solid")

        # Get the indices where the successful_solves mask is False
        unsuccessful_indices: npt.NDArray[np.int_] = np.where(
            np.array(self._solver_status) == False  # noqa: E712
        )[0]

        with pd.ExcelWriter(output_file, engine="openpyxl") as writer:
            for df_name, df in out.items():
                df.to_excel(writer, sheet_name=df_name, index=True)
                sheet = writer.sheets[df_name]

                # Apply highlighting to the rows where the solver failed to find a solution
                for idx in unsuccessful_indices:
                    # Highlight the entire row (starting from index 2 to skip header row)
                    for col in range(1, len(df.columns) + 2):
                        # row=idx+2 because Excel is 1-indexed and row 1 is the header
                        cell = sheet.cell(row=idx + 2, column=col)
                        cell.fill = highlight_fill

        logger.info("Output written to %s", output_file)

    def to_pickle(
        self, file_prefix: Path | str = "new_atmodeller_out", drop_unsuccessful: bool = False
    ) -> None:
        """Writes the output to a pickle file.

        Args:
            file_prefix: Prefix of the output file. Defaults to new_atmodeller_out.
            drop_unsuccessful: Drop models that did not solve. Defaults to False.
        """
        logger.info("Writing output to pickle")
        out: dict[str, pd.DataFrame] = self.to_dataframes(drop_unsuccessful)
        output_file: Path = Path(f"{file_prefix}.pkl")

        with open(output_file, "wb") as handle:
            pickle.dump(out, handle, protocol=pickle.HIGHEST_PROTOCOL)

        logger.info("Output written to %s", output_file)


def collapse_single_entry_values(
    input_dict: dict[str, ArrayLike],
) -> dict[str, ArrayLike]:
    """Collapses single entry values in a dictionary

    Args:
        input_dict: Input dictionary

    Returns:
        Dictionary with collapsed values
    """
    out: dict[str, ArrayLike] = {}
    for key, value in input_dict.items():
        try:
            if value.size > 1:  # type: ignore because AttributeError dealt with
                out[key] = jnp.squeeze(value)
            else:
                out[key] = value.item()  # type:ignore because AttributeError dealt with
        except AttributeError:
            out[key] = value

    return out


def expand_dict(some_dict: dict[str, ArrayLike], expand_to_size: int) -> dict[str, Array]:
    """Gets a dictionary of the values with scalars expanded

    Args:
        some_dict: Some dictionary
        expand_to_size: Size to expand the arrays to

    Returns:
        A dictionary with values expanded to `expand_to_size`
    """

    def expand_to_match_size(x: ArrayLike, size: int) -> ArrayLike:
        """Expands an array

        Args:
            x: Value to possibly expand
            size: Size to expand to

        Returns:
            Expanded value
        """
        if jnp.isscalar(x):
            return jnp.broadcast_to(x, size)

        return x

    expanded_dict: dict[str, Array] = tree_map(
        lambda x: jnp.asarray(expand_to_match_size(x, expand_to_size)), some_dict
    )

    return expanded_dict


def split_dict_by_columns(dict_to_split: dict[str, Array]) -> list[dict[str, Array]]:
    """Splits a dictionary based on columns in the values.

    Args:
        dict_to_split: A dictionary to split

    Returns:
        A list of dictionaries split by column
    """
    # Get the number of columns from the first array in the dictionary
    first_key: str = next(iter(dict_to_split))
    logger.debug("first_key = %s", first_key)
    num_columns: int = dict_to_split[first_key].shape[1]

    # Function to split an array into a list of its columns
    def split_columns(array: Array) -> list[Array]:
        return [array[:, i] for i in range(num_columns)]

    # Apply the splitting function to each array in the dictionary
    split_values: dict[str, list[Array]] = jax.tree_util.tree_map(split_columns, dict_to_split)

    # Initialize a list of dictionaries, one for each column
    split_dicts: list[dict] = [{} for _ in range(num_columns)]

    # Fill the dictionaries with the corresponding columns
    for key, columns in split_values.items():
        for i, column in enumerate(columns):
            split_dicts[i][key] = column

    return split_dicts


def nested_dict_to_dataframes(
    nested_dict: dict[str, dict[str, Any]],
) -> dict[str, pd.DataFrame]:
    """Creates a dictionary of dataframes from a nested dictionary

    Args:
        nested_dict: A nested dictionary

    Returns:
        A dictionary of dataframes
    """
    dataframes: dict[str, pd.DataFrame] = {}

    for outer_key, inner_dict in nested_dict.items():
        # Convert inner dictionary to DataFrame
        df: pd.DataFrame = pd.DataFrame(inner_dict)
        dataframes[outer_key] = df

    return dataframes



#
# Copyright 2024 Dan J. Bower
#
# This file is part of Atmodeller.
#
# Atmodeller is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Atmodeller is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Atmodeller. If not,
# see <https://www.gnu.org/licenses/>.
#
"""Real gas EOS library

Usage:
    from atmodeller.eos.library import get_eos_models
    eos_models = get_eos_models()
    CH4_beattie = eos_models["CH4_beattie_holley58"]
    # Evaluate fugacity at 10 bar and 800 K
    fugacity = CH4_beattie.fugacity(800, 10)
    print(fugacity)
"""

from atmodeller.eos._chabrier import get_chabrier_eos_models
from atmodeller.eos._holland_powell import get_holland_eos_models
from atmodeller.eos._holley import get_holley_eos_models
from atmodeller.eos._saxena import get_saxena_eos_models
from atmodeller.eos._zhang_duan import get_zhang_eos_models
from atmodeller.eos.core import RealGas


def get_eos_models() -> dict[str, RealGas]:
    """Gets a dictionary of EOS models

    Returns:
        Dictionary of EOS models
    """
    # Merge Chabrier models
    eos_models = get_chabrier_eos_models()
    # Merge Holley models
    eos_models |= get_holley_eos_models()
    # Merge Holland and Powell models
    eos_models |= get_holland_eos_models()
    # Merge Saxena models
    eos_models |= get_saxena_eos_models()
    # Merge Zhang models
    eos_models |= get_zhang_eos_models()

    return eos_models
